from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0, qrm1
from plot_util import plot_output

# Do not specify path. This is not a Q1ASM reference test.
instrument = Q1Instrument()
instrument.add_qcm(qcm0)
instrument.add_qrm(qrm1)
instrument.add_control('P1', qrm1.name, [0])
instrument.add_control('P2', qcm0.name, [1])
instrument.add_control('P3', qcm0.name, [2])
instrument.add_control('P4', qcm0.name, [3])
instrument.add_control('R1', qrm1.name, [1])

p = instrument.new_program('sync')
p.repetitions = 10_000_000

t = 5000
a = 1.0

P1 = p.P1
P2 = p.P2

gates=['P1', 'P2', 'P3', 'P4', 'R1']
with p.loop_range(1000):
    p.set_offsets(gates, [-a]*5)
    p.wait(t)
    p.set_offsets(gates, [a]*5)
    p.wait(t)

#p.set_offsets(gates, [0.0]*4)
#p.wait(t)
p.describe()

p.compile(listing=True)

instrument.run_program(p)

plot_output([qcm0, qrm1])



from dataclasses import dataclass

from q1pulse.lang.exceptions import (
        Q1InternalError,
        Q1TimingError
        )


@dataclass
class Instruction:
    mnemonic: str
    args: tuple[int] | None = None
    comment: str | None = None
    label: str | None = None
    wait_after: int | None = None
    overwritten: bool = False


@dataclass
class PendingUpdate:
    KEEP = 0
    MERGE = 1
    FLUSH = 2
    index: int
    time: int


RT_RESOLUTION = 4
MAX_WAIT = (1 << 16) - RT_RESOLUTION


class InstructionQueue:
    _check_time_reg = True
    emulate_signed = True

    def __init__(self, add_comments=False):
        self.add_comments = add_comments
        self._init_section = []
        self._instructions = []
        self._reg_comment = None
        self._wait_loop_cnt = 0
        self._rt_time = 0
        self._pending_update = None
        self._last_rt_command = None
        self._n_rt_instructions = 0
        self._updating_reg = None

    def add_comment(self, line, init_section=False):
        if init_section:
            self._init_section.append(line)
        else:
            self._instructions.append(line)

    def set_label(self, label):
        '''
        Label to be added to next instruction
        '''
        self._instructions.append(Instruction(None, label=label))

    def adjust_time(self, duration):
        self._rt_time += duration

    def __append_instruction(self, instruction):
        self._wait_register_updates(instruction)
        self._instructions.append(instruction)

    def _add_reg_instruction(self, mnemonic, *args, init_section=False):
        if self._reg_comment:
            comment = self._reg_comment
            self._reg_comment = None
        else:
            comment = None
        instruction = Instruction(mnemonic, args, comment=comment)
        if init_section:
            self._init_section.append(instruction)
            return
        self.__append_instruction(instruction)
        self._updating_reg = args[-1]

    def _add_instruction(self, mnemonic, *args, comment=None):
        instruction = Instruction(mnemonic, args, comment=comment)
        self.__append_instruction(instruction)

    def _add_rt_setting(self, mnemonic, *args, time=None):
        if self.add_comments:
            comment = f'@ {time}'
        else:
            comment = None
        instruction = Instruction(mnemonic, args, comment=comment)
        self.__append_instruction(instruction)
        self._schedule_update(time)
        return instruction

    def _add_rt_command(self, mnemonic, *args, time=None, updating=False):
        self._wait_till(time,
                        pending_update=PendingUpdate.MERGE if updating else PendingUpdate.FLUSH)
        wait_after = RT_RESOLUTION
        if self.add_comments:
            comment = f't={time}'
        else:
            comment = None
        instruction = Instruction(mnemonic, args, comment=comment, wait_after=wait_after)
        self.__append_instruction(instruction)
        self._last_rt_command = instruction
        self._n_rt_instructions += 1
        self._rt_time += wait_after

    def _overwrite_rt_setting(self, instruction):
        instruction.overwritten = True
        if self.add_comments:
            instruction.comment += ' = overwritten ='

    def _reset_time(self):
        self._flush_pending_update()
        # start counting after the wait_sync statement
        self._rt_time = 0
        self._last_rt_command = None

    def _schedule_update(self, time):
        self._wait_till(time)
        # put new update immediately after rt_setting (but get index after previous pending has been added)
        index_update = len(self._instructions)
        self._pending_update = PendingUpdate(index_update, time)

    def _wait_till(self, time, pending_update=PendingUpdate.KEEP, return_negative=False):
        if self._pending_update is not None:
            if time == self._pending_update.time:
                if pending_update == PendingUpdate.KEEP:
                    return
                elif pending_update == PendingUpdate.MERGE:
                    # merge pending update with RT command that follows
                    self._pending_update = None
                    self._last_rt_command = None
                    self._rt_time = time
                    return
                elif pending_update != PendingUpdate.FLUSH:
                    Q1InternalError(f'Unknown {pending_update}')
            self._flush_pending_update()
        wait_time = time - self._rt_time
        if wait_time > 0:
            if self._last_rt_command is not None:
                if self._last_rt_command.wait_after + wait_time <= MAX_WAIT:
                    self._last_rt_command.wait_after += wait_time
                else:
                    rem_wait_time = wait_time - MAX_WAIT + self._last_rt_command.wait_after
                    self._last_rt_command.wait_after = MAX_WAIT
                    self.__add_wait_instruction(rem_wait_time)
            else:
                self.__add_wait_instruction(wait_time)
        elif wait_time < 0:
            if return_negative:
                # time will be compensated in register
                self._rt_time = time
                return -wait_time
            raise Q1TimingError(f'Too short wait time of {wait_time} ns at t={self._rt_time}')
        self._last_rt_command = None
        self._rt_time = time
        return 0

    def _flush_pending_update(self):
        pending_update = self._pending_update
        if pending_update is not None:
            wait_after = RT_RESOLUTION
            if self.add_comments:
                comment = f't={pending_update.time}'
            else:
                comment = None
            instruction = Instruction('upd_param',  wait_after=wait_after, comment=comment)
            self._instructions.insert(pending_update.index, instruction)
            self._n_rt_instructions += 1
            self._pending_update = None
            self._last_rt_command = instruction
            self._rt_time += wait_after

    def __add_wait_instruction(self, time):
        if time < MAX_WAIT:
            self._add_instruction('wait', time)
            self._n_rt_instructions += 1
        else:
            n_max, rem_wait = divmod(time, MAX_WAIT)
            if n_max <= 2:
                for _ in range(n_max):
                    self._add_instruction('wait', MAX_WAIT)
            else:
                self._wait_loop_cnt += 1
                with self.temp_regs(1) as wait_reg:
                    self._add_reg_instruction('move', n_max, wait_reg)
                    label = f'wait{self._wait_loop_cnt}'
                    self.set_label(label)
                    self._add_instruction('wait', MAX_WAIT)
                    self._add_instruction('loop', wait_reg, '@'+label)
            self._n_rt_instructions += n_max
            if rem_wait > 0:
                self._add_instruction('wait', rem_wait)
                self._n_rt_instructions += 1

    def _add_wait_reg(self, time_reg, elapsed=0, less_then_65us=False):  # @@@ make use of option less_then_65us
        self._n_rt_instructions += 1  # this is not correct if > 65 us.
        if less_then_65us and elapsed == 0 and not self._check_time_reg:
            # single instruction for short simple wait
            self._add_instruction('wait', time_reg)
            return

        wait_reg = self.allocate_reg('_waittime')
        if elapsed > 0:
            self._add_reg_instruction('sub', time_reg, elapsed, wait_reg)
        else:
            self._add_reg_instruction('move', time_reg, wait_reg)
        self._wait_loop_cnt += 1

        if self._check_time_reg:
            self.add_comment('         --- check for negative wait time')
            continue_label = f'waitc{self._wait_loop_cnt}'
            if self.emulate_signed:
                self.add_comment('         --- emulate signed wait time')
                with self.temp_regs(1) as temp_reg:
                    self._add_reg_instruction('xor', wait_reg, 0x8000_0000, temp_reg)
                    self._add_instruction('jge', temp_reg, 0x8000_0000 + RT_RESOLUTION, '@'+continue_label)
            else:
                self._add_instruction('jge', wait_reg, RT_RESOLUTION, '@'+continue_label)
            self._add_instruction('illegal', comment='negative wait time')
            if less_then_65us:
                self._add_instruction('jlt', wait_reg, MAX_WAIT, '@'+continue_label)
                self._add_instruction('illegal', comment='larger than 65 us')
            self.set_label(continue_label)

        if not less_then_65us:
            loop_label = f'wait{self._wait_loop_cnt}'
            end_label = f'endwait{self._wait_loop_cnt}'
            self._add_instruction('jlt', wait_reg, MAX_WAIT, '@'+end_label)
            self.set_label(loop_label)
            self._add_instruction('wait', MAX_WAIT-RT_RESOLUTION)
            self._add_reg_instruction('sub', wait_reg, MAX_WAIT-RT_RESOLUTION, wait_reg)
            self._add_instruction('jge', wait_reg, MAX_WAIT, '@'+loop_label)
            self.set_label(end_label)

        self._add_instruction('wait', wait_reg)

    def _wait_register_updates(self, instruction):
        '''
        Inserts a NOP when one of the arguments is a register which
        has been modified by the previous instruction.
        Registers are modified by the arithmetic instructions.
        '''
        if self._updating_reg is not None and self._updating_reg in instruction.args:
            self._add_instruction('nop',
                                  comment=f' {instruction.mnemonic} wait for {self._updating_reg}')
        self._updating_reg = None



import numpy as np
import matplotlib.pyplot as pt
from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0, qrm1
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_qrm(qrm1)
instrument.add_control('q1', qcm0.name, [0,1], nco_frequency=200e6)
instrument.add_control('q2', qrm1.name, [0,1], nco_frequency=200e6)
instrument.add_control('P1', qcm0.name, [2])
instrument.add_control('P2', qcm0.name, [3])
instrument.add_readout('R1', qrm1.name, [], nco_frequency=200e6)

qrm1.in0_gain(0)
qrm1.in1_gain(0)

p = instrument.new_program('nco_sync')
p.repetitions = 1

q1 = p.q1
q2 = p.q2
P1 = p.P1
P2 = p.P2
R1 = p['R1']

N = 10000

R1.add_acquisition_bins('default', N)
R1.integration_length_acq = 400

with p.loop_range(N):

    with p.parallel():
#        q1.block_pulse(600, 0.1)
        q2.block_pulse(600, 0.5)
        R1.acquire('default', 'increment', t_offset=160)
        p.wait(1000)
    p.wait(50_000-1620)

#p.describe()

p.compile(listing=True)

instrument.run_program(p)

plot_output([qcm0, qrm1])

data = instrument.get_acquisition_bins('R1', 'default')

#pprint(data)
I = np.array(data['integration']['path0'])/R1.integration_length_acq
Q = np.array(data['integration']['path1'])/R1.integration_length_acq
c = I + 1j*Q
#for i in range(N):
#    print(f'{I[i]:6.3f}, {Q[i]:6.3f}, {abs(c[i]):6.3f}, {np.angle(c[i])/np.pi:5.3f}')

pt.figure()
pt.plot(np.angle(c)/np.pi)
pt.title('phase')

pt.figure()
pt.plot(np.abs(c))
pt.title('amplitude')



import logging
from dataclasses import dataclass
from abc import abstractmethod

from .sequencer_states import translate_seq_status, translate_seq_state
from q1pulse.util.qblox_version import qblox_version, Version


logger = logging.getLogger(__name__)


@dataclass
class Sequencer:
    module_name: str
    seq_nr: int
    channels: list[int]
    enabled_paths: list[int]
    max_output_voltage: float
    nco_frequency: float | None = None
    in_channels: list[int] | None = None
    label: str | None = None


class QbloxModule:
    verbose = False
    n_sequencers = 6

    def __init__(self, pulsar):
        self.name = pulsar.name
        # check module is present in slot.
        if hasattr(pulsar, 'present'):
            if not pulsar.present():
                raise Exception('No module in slot {pulsar.slot_idx}')
        self.pulsar = pulsar
        self._allocated_seq = 0
        self.disable_all_out()
        # disable all sequencers
        for seq_nr in range(0, self.n_sequencers):
            self.enable_sync(seq_nr, False)

    def get_sequencer(self, channels):
        seq_nr = self._allocate_seq_number()
        seq_paths = self._get_seq_paths(channels)
        return Sequencer(self.name, seq_nr, channels, seq_paths,
                         self.max_output_voltage)

    def _allocate_seq_number(self):
        if self._allocated_seq == self.n_sequencers:
            raise Exception(f'Module {self.name} is out of sequencers. '
                            f'All {self.n_sequencers} are already allocated.')
        sequencer_nr = self._allocated_seq
        self._allocated_seq += 1
        return sequencer_nr

    @abstractmethod
    def _get_seq_paths(self, channels):
        pass

    def disable_all_out(self):
        for seq_nr in range(0, self.n_sequencers):
            # Note: RF module connect outputs in pairs.
            n_out_ch = self.n_channels // 2 if self.pulsar.is_rf_type else self.n_channels
            for out in range(0, n_out_ch):
                self._sset(seq_nr, f'connect_out{out}', 'off')

    def set_label(self, seq_nr, label):
        self.pulsar.sequencers[seq_nr].label = label

    def upload(self, seq_nr, sequence):
        if isinstance(sequence, str):
            # don't cache sequence. The contents of the file might have changed.
            filename = sequence
            # print(f'Loading {filename} to sequencer {self.pulsar.name}:{seq_nr}')
            self._sset(seq_nr, 'sequence', filename, cache=False)
        else:
            self._sset(seq_nr, 'sequence', sequence)

    def arm_sequencers(self):
        for seq_nr in range(0, self.n_sequencers):
            if self.enabled(seq_nr):
                self.pulsar.arm_sequencer(seq_nr)

    def start_sequencers(self):
        enabled = [self.enabled(seq_nr) for seq_nr in range(0, self.n_sequencers)]
        if any(enabled):
            self.pulsar.start_sequencer()

    def stop_sequencers(self):
        self.pulsar.stop_sequencer()

    def get_sequencer_status(self, seq_nr, timeout=0):
        if qblox_version >= Version('0.12.0'):
            status = self.pulsar.get_sequencer_status(seq_nr, timeout)
            return translate_seq_status(status)
        else:
            state = self.pulsar.get_sequencer_state(seq_nr, timeout)
            return translate_seq_state(state)

    def enable_sync(self, seq_nr, enable):
        self._sset(seq_nr, 'sync_en', enable)

    def enable_out(self, seq_nr, channel):
        # Keep old convention: I on 0 and 2, Q on 1 and 3
        # TODO: change API to allow different IQ mapping.
        value = 'I' if channel  % 2 == 0 else 'Q'
        self._sset(seq_nr, f'connect_out{channel}', value)

    def enable_out_iq(self, seq_nr, channels):
        if len(channels) == 0:
            return
        if len(channels) != 2 or channels[0] // 2 != channels[1] // 2:
            raise Exception(f'Incorrect channels {channels}. RF output must be enabled in pairs')
        ch = channels[0] // 2
        self._sset(seq_nr, f'connect_out{ch}', 'IQ')

    def set_nco(self, seq_nr, nco_frequency):
        self._sset(seq_nr, 'mod_en_awg', nco_frequency is not None)
        if nco_frequency is not None:
            self._sset(seq_nr, 'nco_freq', nco_frequency)

    def set_mixer_gain_ratio(self, seq_nr, value):
        self._sset(seq_nr, 'mixer_corr_gain_ratio', value)

    def set_mixer_phase_offset_degree(self, seq_nr, value):
        self._sset(seq_nr, 'mixer_corr_phase_offset_degree', value)

    def configure_trigger_counter(self, seq_nr, address, threshold, invert):
        self._sset(seq_nr, f'trigger{address}_count_threshold', threshold)
        self._sset(seq_nr, f'trigger{address}_threshold_invert', invert)

    def set_awg_offsets(self, seq_nr, offset0, offset1):
        self._sset(seq_nr, 'offset_awg_path0', offset0)
        self._sset(seq_nr, 'offset_awg_path1', offset1)

    def enabled(self, seq_nr):
        seq = getattr(self.pulsar, f'sequencer{seq_nr}')
        param = seq.parameters['sync_en']
        return param.cache()

    def disable_seq(self, sequencer):
        seq_nr = sequencer.seq_nr
        self.enable_sync(seq_nr, False)

    def enable_seq(self, sequencer):
        seq_nr = sequencer.seq_nr
        self.enable_sync(seq_nr, True)
        if self.pulsar.is_rf_type:
            self.enable_out_iq(seq_nr, sequencer.channels)
        else:
            for ch in sequencer.channels:
                self.enable_out(seq_nr, ch)

    def _sset(self, seq_nr, name, value, cache=True):
        full_name = f'sequencer{seq_nr}.{name}'
        seq = getattr(self.pulsar, f'sequencer{seq_nr}')
        param = seq.parameters[name]
        try:
            if cache and param.cache.valid and param.cache() == value:
                if QbloxModule.verbose:
                    logger.debug(f'# {full_name}={value} -- cached')
                return
        except:
            logger.debug(f'No cache value for {full_name}')
        result = param(value)
        if QbloxModule.verbose:
            logger.info(f'{full_name}={value}')
        return result

    def invalidate_cache(self, seq_nr, param_name):
        seq = getattr(self.pulsar, f'sequencer{seq_nr}')
        param = seq.parameters[param_name]
        param.cache.invalidate()

    def set_out_offset(self, channel, offset_mV):
        if self.pulsar.is_rf_type:
            name = f'out{channel//2}_offset_path{channel%2}'
            value = offset_mV
        else:
            name = f'out{channel}_offset'
            value = offset_mV/1000

        param = self.pulsar.parameters[name]
        if param.cache() == value:
            if QbloxModule.verbose:
                logger.debug(f'# {name}={value} -- cached')
            return
        param(value)

    def set_marker_invert(self, number: int, invert: bool):
        name = f'marker{number}_inv_en'
        param = self.pulsar.parameters[name]
        if param.cache() != invert:
            param(invert)


class QcmModule(QbloxModule):
    n_channels = 4

    def __init__(self, pulsar):
        super().__init__(pulsar)
        self.max_output_voltage = 2.5 if not pulsar.is_rf_type else 3.3

    def _get_seq_paths(self, channels):
        if len(channels) == 0:
            # no output, only markers
            seq_channels = []

        elif len(channels) == 1:
            channel = channels[0]
            if channel in [0, 1]:
                seq_channels = [channel]
            elif channel in [2, 3]:
                seq_channels = [channel - 2]
            else:
                raise Exception(f'illegal channel combination {channels}')

        elif len(channels) == 2:
            channel_pair = sorted(channels)
            if channel_pair == [0, 1]:
                seq_channels = channels
            elif channel_pair == [2, 3]:
                seq_channels = [ch-2 for ch in channels]
            else:
                raise Exception(f'illegal channel combination {channels}')
        else:
            raise Exception(f'illegal channel combination {channels}')

        return seq_channels


class QrmModule(QbloxModule):
    n_channels = 2

    def __init__(self, pulsar):
        super().__init__(pulsar)
        self.max_output_voltage = 0.5 if not pulsar.is_rf_type else 3.3
        self.disable_all_inputs()

    def _get_seq_paths(self, channels):
        for channel in channels:
            if channel not in [0, 1]:
                raise Exception(f'illegal channel combination {channels}')

        return channels

    def disable_all_inputs(self):
        for seq_nr in range(0, self.n_sequencers):
            if self.pulsar.is_rf_type:
                self._sset(seq_nr, 'connect_acq', 'off')
            else:
                self._sset(seq_nr, 'connect_acq_I', 'off')
                self._sset(seq_nr, 'connect_acq_Q', 'off')

    def thresholded_acq_rotation(self, seq_nr, phase_rotation):
        self._sset(seq_nr, 'thresholded_acq_rotation', phase_rotation)

    def thresholded_acq_threshold(self, seq_nr, threshold):
        self._sset(seq_nr, 'thresholded_acq_threshold', threshold)

    def integration_length_acq(self, seq_nr, length):
        self._sset(seq_nr, 'integration_length_acq', length)

    def nco_prop_delay(self, seq_nr, delay):
        if delay > 0:
            if 96 <= delay <= 245:
                self._sset(seq_nr, 'nco_prop_delay_comp_en', True)
                self._sset(seq_nr, 'nco_prop_delay_comp', delay - 146)
            else:
                logger.warning(f"NCO delay ({delay} ns) is out of range. NCO delay not enabled.")
                self._sset(seq_nr, 'nco_prop_delay_comp_en', False)
        else:
            self._sset(seq_nr, 'nco_prop_delay_comp_en', False)

    def delete_acquisition_data(self, seq_nr):
        self.pulsar.delete_acquisition_data(seq_nr, all=True)

    def enable_in(self, seq_nr, channel):
        if self.pulsar.is_rf_type:
            self._sset(seq_nr, 'connect_acq', 'in0')
        else:
            # Keep old convention: I on 0, Q on 1
            # TODO: change API to allow different IQ mapping.
            if channel == 0:
                self._sset(seq_nr, 'connect_acq_I', 'in0')
            else:
                self._sset(seq_nr, 'connect_acq_Q', 'in1')

    def enable_seq(self, sequencer):
        super().enable_seq(sequencer)
        if sequencer.in_channels is not None:
            seq_nr = sequencer.seq_nr
            for ch in sequencer.in_channels:
                self.enable_in(seq_nr, ch)

    def set_nco(self, seq_nr, nco_frequency):
        super().set_nco(seq_nr, nco_frequency)
        self._sset(seq_nr, 'demod_en_acq', nco_frequency is not None)

    def set_trigger(self, seq_nr, address, invert=False):
        enabled = address is not None and address > 0
        self._sset(seq_nr, 'thresholded_acq_trigger_en', enabled)
        if enabled:
            self._sset(seq_nr, 'thresholded_acq_trigger_address', address)
            self._sset(seq_nr, 'thresholded_acq_trigger_invert', invert)




from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_control('P1', qcm0.name, [2])

p = instrument.new_program('registers')

seq = p.P1

# just add some waiting time to prevent error messages on real-time executor
p.wait(1000)
p.P1.set_offset(0.0)


p.R.a = 10

p.R.b = p.R.a + 1
p.R.b = 5 + p.R.a + 1

p.R.c = p.R.b + p.R.a
p.R.c += 5
p.R.c -= 17
p.R.d = 1 - p.R.a

p.R.e = p.R.a << 2
p.R.f = p.R.a >> 2
p.R.h = p.R.d >> 2


# floating point
p.R.x = 1.0
p.R.x -= 0.8
p.R.y = 0.5
p.R.z = p.R.x - p.R.y

seq.Rs.x = 9
seq.Rs.y = seq.Rs.x + p.R.b

# bitwise
seq.Rs.a = 0b0011
seq.Rs.b = seq.Rs.a & 0b0101
seq.Rs.c = seq.Rs.a | 0b0101
seq.Rs.d = 0b0101 ^ seq.Rs.a
seq.Rs.e = ~seq.Rs.a

seq.describe()

p.compile(annotate=True, listing=True, optimize=0)

instrument.run_program(p)

if hasattr(qcm0, 'print_registers'):
    # get result from Q1Simulator
    qcm0.print_registers(0, range(20))





from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_control('P1', qcm0.name, [2])
instrument.add_control('P2', qcm0.name, [3])

p = instrument.new_program('loop_range')
p.repetitions = 3

P1 = p.P1
P2 = p.P2

P1.block_pulse(200, -0.1)

with p.loop_range(20, 100, 20) as t_wait:
    P1.block_pulse(20, 0.8)
    p.wait(t_wait)
    P2.block_pulse(20, 0.6)
    p.wait(200)

# 4 pulses
with p.loop_range(-2, 2, 1):
    P1.block_pulse(20, 0.2)
    p.wait(100)

p.wait(100)

# 4 pulses
with p.loop_range(80, 0, -20) as t_wait:
    P1.block_pulse(20, 0.1)
    p.wait(t_wait + 20)

#p.describe()
#print()

p.compile(listing=True, annotate=True)

instrument.run_program(p)

plot_output([qcm0])



from .register_statements import RegisterAssignment
from .math_expressions import Operand

class Register(Operand):
    def __init__(self, name, local=False, dtype=None):
        self.name = f'Rs.{name}' if local else f'R.{name}'
        self._initialized = False
        self._dtype = dtype

    @property
    def dtype(self):
        return self._dtype

    def assign(self, value_or_expression, allocate=False):
        allocate = not self._initialized or allocate
        if not self._initialized:
            self._initialized = True
        return RegisterAssignment(self, value_or_expression, allocate=allocate)

    def __repr__(self):
        return self.name



from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_control('P1', qcm0.name, [2])

p = instrument.new_program('asr')

seq = p.P1

p.R.a = 9
p.R.b = 2

p.R.c = p.R.a >> 2

p.R.d = p.R.a >> p.R.b

p.R.e = -9
p.R.f = p.R.e >> 1  # q1asm evaluation
p.R.h = -9 >> 1     # Python evaluation

seq.describe()

p.compile(annotate=True, listing=True, optimize=0)

instrument.run_program(p)

if hasattr(qcm0, 'print_registers'):
    # get result from Q1Simulator
    qcm0.print_registers(0, range(10))




from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0, qrm1
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_qrm(qrm1)
instrument.add_control('q1', qcm0.name, [0,1], nco_frequency=50e6)
instrument.add_control('P1', qcm0.name, [2])
instrument.add_control('P2', qcm0.name, [3])
instrument.add_readout('R1', qrm1.name, [1])

p = instrument.new_program('ramsey')
p.repetitions = 2

q1 = p.q1
P1 = p.P1
P2 = p.P2
R1 = p['R1']

R1.add_acquisition_bins('q1value', 20)

gates=['P1', 'P2']
v_init = [0.120, 0.040]
v_manip = [0.0, 0.0]
v_read = [-0.030, 0.060]
t_pulse = 80
v_pulse = 0.5

R1.reset_bin_counter('q1value')
with p.loop_range(100, 2001, 100) as t_wait:
    p.block_pulse(200, gates, v_init)
    p.ramp(20, gates, v_init, v_manip)

    q1.block_pulse(t_pulse, v_pulse)
    p.wait(t_wait)
    q1.block_pulse(t_pulse, v_pulse)

    p.ramp(100, gates, v_manip, v_read)
    with p.parallel():
        p.set_offsets(gates, v_read)
        p.wait(1000)
        R1.acquire('q1value', 'increment', t_offset=100)
    p.set_offsets(gates, [0.0, 0.0])

p.describe()

p.compile(listing=True, annotate=True)

instrument.run_program(p)

plot_output([qcm0])

if hasattr(qrm1, 'print_acquisitions'):
    # get acquisition timing and counts from Q1Simulator
    qrm1.print_acquisitions()



from contextlib import contextmanager
from ..lang.exceptions import Q1NameError, Q1MemoryError

class SequencerRegisters:
    stack_size = 32

    def __init__(self, log_func):
        super().__init__()
        self._log = log_func
        self._allocated_regs = {}
        # stack for registers allcated in scope
        self._stack_ptr = 0
        self._scope = []
        self.enter_scope()

    def get_asm_reg(self, name):
        try:
            return self._allocated_regs[name]
        except KeyError:
            raise Q1NameError(f'Register {name} not defined')

    def allocate_reg(self, name, log=True):
        if name in self._allocated_regs:
            self._print_reg_admin()
            raise Q1NameError(f'Register {name} already allocated')
        try:
            asm_reg = self._allocate_stack_reg(name)
            self._allocated_regs[name] = asm_reg
            if log and self._log is not None:
                self._log(f'{asm_reg}: {name}')
            return asm_reg
        except IndexError:
            raise Q1MemoryError(f'Cannot allocate register {name}')

    def enter_scope(self):
        self._scope.append((self._stack_ptr, {}))

    def exit_scope(self):
        ptr, named = self._scope.pop()
        self._stack_ptr = ptr
        for reg_name in named:
            del self._allocated_regs[reg_name]

    @contextmanager
    def temp_regs(self, n):
        self.enter_scope()
        regs = [self.get_temp_reg(log=False) for i in range(n)]
        if self._log is not None:
            self._log(f'temp {regs}')
        yield regs if n > 1 else regs[0]
        self.exit_scope()

    def _allocate_stack_reg(self, name=None):
        reg_nr = self._stack_ptr
        self._stack_ptr += 1
        if self._stack_ptr >= SequencerRegisters.stack_size:
            raise Q1MemoryError('Stack overflow')
        reg_name = f'R{reg_nr}'
        if name:
            self._scope[-1][1][name] = reg_name
        return reg_name

    def get_temp_reg(self, log=True):
        asm_reg = self._allocate_stack_reg()
        if log and self._log:
            self._log(f'temp {asm_reg}')
        return asm_reg

    def _print_reg_admin(self):
        # print(f'Free: {self._free_regs}')
        reg_names = {f'R{i}':name for name,i in self._allocated_regs.items()}
        print('Allocated:')
        for r,name in sorted(reg_names.items()):
            print(f'  {r}: {name}')
        indent = 0
        for scope in self._scope_regs:
            indent += 4
            spaces = ' '*indent
            print(f'{spaces}> {scope}')




import scipy.signal as signal

from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0, qrm1
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_qrm(qrm1)
instrument.add_control('q1', qcm0.name, [0,1], nco_frequency=100e6)
instrument.add_control('P1', qcm0.name, [2])
instrument.add_control('P2', qcm0.name, [3])
instrument.add_readout('R1', qrm1.name, [])

p = instrument.new_program('shaped')
p.repetitions = 2

q1 = p.q1
P1 = p.P1
P2 = p.P2
R1 = p.R1


gauss80 = q1.add_wave('gauss80', signal.windows.gaussian(80, std=0.12 * 80))
tukey100 = P1.add_wave('tukey100', signal.windows.tukey(100, 0.5))

R1.add_acquisition_bins('acq_bins', p.repetitions)
R1.integration_length_acq = 80

amplitude = 0.125

q1.shaped_pulse(gauss80, 0.4, gauss80, 0.4)
p.wait(20)
P1.shaped_pulse(tukey100, 0.1)
p.wait(20)
q1.shaped_pulse('gauss80', amplitude, 'gauss80', amplitude)

p.wait(40)
with p.parallel():
    P1.block_pulse(100, 0.25)
    P2.block_pulse(100, -0.25)
    q1.shaped_pulse(gauss80, amplitude)
    R1.acquire('acq_bins', 'increment')

p.describe()
print()

p.compile(listing=True, annotate=True)

instrument.run_program(p)

plot_output([qcm0, qrm1])

if hasattr(qrm1, 'print_acquisitions'):
    # get acquisition timing and counts from Q1Simulator
    qrm1.print_acquisitions()



import numpy as np
import scipy.signal as signal

from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0, qrm1
from plot_util import plot_output

def set_mock_data(qrm, seq_nr, acq_name, data):
    seq = qrm.sequencers[seq_nr]
    if hasattr(seq, 'set_acquisition_mock_data'):
        seq.set_acquisition_mock_data([data], name=acq_name, repeat=True)

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_qrm(qrm1)
instrument.add_control('P1', qcm0.name, [0])
instrument.add_control('P2', qrm1.name, [1])
instrument.add_control('P2c', qcm0.name, [1])
instrument.add_readout('R1', qrm1.name, [])

# gain = 0: +/- 0.5V; gain -6 dB: +/- 1.0 V
vmax_in = 0.5
qrm1.in0_gain(0)
qrm1.in1_gain(0)

p = instrument.new_program('acquire_mocked')
p.repetitions = 1

P1 = p.P1
P2 = p.P2
P2c = p.P2c
R1 = p.R1

N = 5
n_acq = N*N*p.repetitions
R1.add_acquisition_bins('non-weighed', n_acq)
R1.add_acquisition_bins('weighed', n_acq)
R1.add_weight('gaus100', signal.windows.gaussian(100, 12))
#R1.add_weight('gaus100', np.ones(100))
R1.integration_length_acq = 100


# output range QCM (P1): +/- 2.5 V
# output range QRM (P1): +/- 0.5 V
# input range QRM (R1): +/- 0.5 V (gain = 0 dB)

v1_max = 0.4/2.5
v2_max = 0.4/0.5

with p.loop_linspace(-v1_max, v1_max, N) as v1:
    with p.loop_linspace(-v2_max, v2_max, N) as v2:
        with p.parallel():
            P1.block_pulse(500, v1)
            P2.block_pulse(500, v2)
            P2c.block_pulse(500, v1)
            # delay from output to input is ~108 ns
            R1.acquire('non-weighed', 'increment', t_offset=160)

        with p.parallel():
            P1.block_pulse(500, v1)
            P2c.block_pulse(500, v1)
            P2.block_pulse(500, v2)
            R1.acquire_weighed('weighed', 'increment', 'gaus100', t_offset=160)

        p.wait(1100)

#p.describe()
#print()

p.compile(listing=True, annotate=True)

#%% Set Mock data
data0 = np.linspace(-v1_max, v1_max, N).repeat(N)
data1 = np.tile(np.linspace(-v2_max, v2_max, N), N)

data0 = np.tile(data0, p.repetitions)
data1 = np.tile(data1, p.repetitions)

set_mock_data(qrm1, 1, 'non-weighed',
              np.array([data0, data1]).T * R1.integration_length_acq)
set_mock_data(qrm1, 1, 'weighed', (data0 + 1j*data1)/2)

#%% Run and plot
# run and get results
instrument.run_program(p)

plot_output([qcm0, qrm1])

# @@@ program should return data.
data_n = instrument.get_acquisition_bins('R1', 'non-weighed')
data_w = instrument.get_acquisition_bins('R1', 'weighed')

dn0 = np.array(data_n['integration']['path0']).reshape((p.repetitions,N,N))/R1.integration_length_acq*vmax_in
dn1 = np.array(data_n['integration']['path1']).reshape((p.repetitions,N,N))/R1.integration_length_acq*vmax_in
dw0 = np.array(data_w['integration']['path0']).reshape((p.repetitions,N,N))*vmax_in
dw1 = np.array(data_w['integration']['path1']).reshape((p.repetitions,N,N))*vmax_in

with np.printoptions(precision=2, threshold=1000):
    print('non-weighed')
    print(dn0)
    print(dn1)
    print('weighed')
    print(dw0)
    print(dw1)



from dataclasses import dataclass


@dataclass
class Trigger:
    sequencer_name: str
    address: int
    ''' range: 1 to 15 '''
    invert: bool = False

@dataclass
class TriggerCounter:
    trigger: Trigger
    threshold: int
    invert: bool = False

    @property
    def address(self):
        return self.trigger.address

    @property
    def name(self):
        return self.trigger.sequencer_name




from .sequencer import SequenceBuilder
from .sequencer_data import Wave, WaveCollection
from ..lang.exceptions import Q1ValueError, Q1TypeError
from ..lang.math_expressions import Expression
from ..lang.register import Register
from ..lang.timed_statements import (
        SetMarkersStatement,
        AwgDcOffsetStatement, AwgGainStatement,
        ShiftPhaseStatement, SetPhaseStatement,
        ResetPhaseStatement,
        PlayWaveStatement, SetFrequencyStatement,
        )


class ControlBuilder(SequenceBuilder):
    def __init__(self, name, enabled_paths, max_output_voltage,
                 nco_frequency=None):
        super().__init__(name)
        self._enabled_paths = enabled_paths
        self.max_output_voltage = max_output_voltage
        self._nco_frequency = nco_frequency
        self._waves = WaveCollection()
        self._mixer_gain_ratio = None
        self._mixer_phase_offset_degree = None

    @property
    def nco_frequency(self):
        return self._nco_frequency

    @nco_frequency.setter
    def nco_frequency(self, value):
        self._nco_frequency = value

    @property
    def mixer_gain_ratio(self):
        return self._mixer_gain_ratio

    @mixer_gain_ratio.setter
    def mixer_gain_ratio(self, value):
        self._mixer_gain_ratio = value

    @property
    def mixer_phase_offset_degree(self):
        return self._mixer_phase_offset_degree

    @mixer_phase_offset_degree.setter
    def mixer_phase_offset_degree(self, value):
        self._mixer_phase_offset_degree = value

    def add_wave(self, name, data):
        return self._waves.add_wave(name, data)

    def set_markers(self, value, t_offset=0):
        t1 = self.current_time + t_offset
        self.set_pulse_end(t1)
        self._add_statement(SetMarkersStatement(t1, value))

    def set_offset(self, value0, value1=None, t_offset=0):
        value0, value1 = self._apply_paths(value0, value1)
        t1 = self.current_time + t_offset
        self.set_pulse_end(t1)
        self._add_statement(AwgDcOffsetStatement(t1, value0, value1))

    def set_gain(self, value0, value1=None, t_offset=0):
        value0, value1 = self._apply_paths(value0, value1)
        t1 = self.current_time + t_offset
        self.set_pulse_end(t1)
        self._add_statement(AwgGainStatement(t1, value0, value1))

    def play(self, wave0, wave1=None, t_offset=0):
        wave0, wave1 = self._apply_paths(wave0, wave1)
        t1 = self.current_time + t_offset
        self.set_pulse_end(t1)
        wave0 = self._translate_wave(wave0)
        wave1 = self._translate_wave(wave1)
        self._add_statement(PlayWaveStatement(t1, wave0, wave1))

    def set_frequency(self, frequency, t_offset=0):
        '''
        Args:
            frequency (float): nco frequency in Hz
        '''
        t1 = self.current_time + t_offset
        self.set_pulse_end(t1)
        self._add_statement(SetFrequencyStatement(t1, frequency))

    def reset_phase(self, t_offset=0):
        t1 = self.current_time + t_offset
        self.set_pulse_end(t1)
        self._add_statement(ResetPhaseStatement(t1))


    def shift_phase(self, delta, t_offset=0, hires_reg=True):
        '''
        Args:
            hires_reg:
                Convert register to exactly phase if True, else
                convert register value with relative error of 2e-4.

        NOTE:
            When `phase` is a Register many instructions are added for the
            conversion of the `phase` This costs ~188 ns when hires_reg=False.
            When hires_reg=True this costs ~268 ns.

            When `phase` is a constant in q1asm, then the resolution is 1e-9.
            No extra time is needed.
        '''
        t1 = self.current_time + t_offset
        self.set_pulse_end(t1)
        # WORKAROUND: Qblox sequencer inverts phase delta when frequency is negative
        # TODO: remove workaround when fixed in firmware
        if self._nco_frequency and self._nco_frequency < 0:
            delta = -delta
        self._add_statement(ShiftPhaseStatement(t1, delta, hires_reg))

    def set_phase(self, phase, t_offset=0, hires_reg=True):
        '''
        Args:
            hires_reg:
                Convert register to exactly phase if True, else
                convert register value with relative error of 2e-4.

        NOTE:
            When `phase` is a Register many instructions are added for the
            conversion of the `phase` This costs ~188 ns when hires_reg=False.
            When hires_reg=True this costs ~268 ns.
        '''
        t1 = self.current_time + t_offset
        self.set_pulse_end(t1)
        self._add_statement(SetPhaseStatement(t1, phase, hires_reg))

    def block_pulse(self, duration, amplitude0, amplitude1=None, t_offset=0):
        self.add_comment(f'block_pulse({duration}, {amplitude0}, {amplitude1})')
        if not isinstance(duration, (Register, Expression)):
            with self._local_timeline(t_offset=t_offset, duration=duration):
                self.set_offset(amplitude0, amplitude1)
                self.wait(duration)
                self.set_offset(0.0, None)
        else:
            # TODO: try to make this cleaner and more flexible.
            if not self._timeline.is_running:
                raise Q1ValueError('Variable pulse length not possible in parallel section')
            self.set_offset(amplitude0, amplitude1)
            self._program.wait(duration)
            self.set_offset(0.0, None)

    def shaped_pulse(self, wave0, amplitude0=None, wave1=None, amplitude1=None, t_offset=0):
        '''
        When amplitude is None, it is assumed that the gain has already been set to
        the right value.
        '''
        wave0 = self._translate_wave(wave0)
        wave1 = self._translate_wave(wave1)
        wave0_name = wave0.name if wave0 is not None else None
        wave1_name = wave1.name if wave1 is not None else None
        self.add_comment(f'shaped_pulse({wave0_name}, {amplitude0}, {wave1_name}, {amplitude1})')

        duration = max(
                len(wave0.data) if wave0 is not None else 0,
                len(wave1.data) if wave1 is not None else 0
                )
        with self._local_timeline(t_offset=t_offset, duration=duration):
            if amplitude0 is not None:
                self.set_gain(amplitude0, amplitude1)
            self.play(wave0, wave1)
            self.wait(duration)

    def ramp(self, duration, v_start, v_end, v_after=0.0, t_offset=0):
        if isinstance(duration, (Register, Expression)):
            raise Q1TypeError('Ramp duration cannot be a variable or expression; '
                              'Unroll loop using Python for-loop.')

        ramp_loop_time = 100
        self.add_comment(f'ramp({duration}, {v_start}, {v_end})')
        with self._local_timeline(t_offset=t_offset, duration=duration):
            if duration <= ramp_loop_time:
                # w_ramp is a wave from 0 to 1.0
                w_ramp = self._waves.get_ramp(duration)
                self.set_gain(v_end - v_start)
                self.set_offset(v_start)
                self.play(w_ramp)
                self.wait(duration)
            elif (isinstance(v_start, (Register, Expression))
                  or isinstance(v_end, (Register, Expression))):
                # divide duration till smallest multiple of 4 larger than or equal to 100
                shift = 0
                wave_duration = duration
                while wave_duration > 200 and wave_duration % 8 == 0:
                    wave_duration >>=1
                    shift += 1
                self.Rs._ramp_step = (v_end - v_start)
                if shift >= 1:
                    self.Rs._ramp_step >>= shift
                # w_ramp is a wave from 0 to 1.0
                w_ramp = self._waves.get_ramp(wave_duration)
                self.Rs._ramp_offset = v_start
                self.set_gain(self.Rs._ramp_step)
                with self._seq_repeat(1 << shift):
                    self.set_offset(self.Rs._ramp_offset)
                    self.play(w_ramp)
                    self.Rs._ramp_offset += self.Rs._ramp_step
                    self.wait(wave_duration)
            else:
                step = (v_end - v_start) * ramp_loop_time / duration
                # step is a fixed point value in q1asm.  Resolution is 1/65536 of LSB output.
                lsb = 1/(2**15)
                if abs(step) > lsb:
                    gain = step
                    min_lsb_steps = 2**6
                    if abs(step) < lsb*min_lsb_steps:
                        # decrease ramp, increase gain for better accuracy
                        w_ramp = self._waves.get_ramp(ramp_loop_time, stop=1/min_lsb_steps)
                        gain = step * min_lsb_steps
                    else:
                        w_ramp = self._waves.get_ramp(ramp_loop_time)
                    # divmod, but with rem [1,100], and n > 1
                    n, rem = divmod(duration-1, ramp_loop_time)
                    rem += 1
                    self.Rs._ramp_offset = v_start
                    self.set_gain(gain)
                    with self._seq_repeat(n):
                        self.set_offset(self.Rs._ramp_offset)
                        self.play(w_ramp)
                        self.Rs._ramp_offset += step
                        self.wait(ramp_loop_time)
                    self.set_offset(self.Rs._ramp_offset)
                    self.play(w_ramp)
                    self.wait(rem)
                    self.set_gain(0.0)
                else:
                    # steps of 1 LSB
                    min_step = lsb
                    n_steps = int(abs(v_end - v_start) / min_step)
                    if n_steps == 0:
                        n = 0
                        rem = duration
                    else:
                        # minimum time multiple of 4 ns
                        t_step = int(duration / n_steps)
                        t_step = int(t_step / 4) * 4
                        # divmod, but with rem [1,100]
                        n, rem = divmod(duration-1, t_step)
                        rem += 1
                        step = (v_end - v_start) * t_step / duration
                    self.Rs._ramp_offset = v_start
                    if n > 0:
                        with self._seq_repeat(n):
                            self.set_offset(self.Rs._ramp_offset)
                            self.Rs._ramp_offset += step
                            self.wait(t_step)
                    self.set_offset(self.Rs._ramp_offset)
                    self.wait(rem)

            if v_after is not None:
                # set constant value
                self.set_offset(v_after)
            else:
                # assume there will be another instruction setting offset
                pass

    def chirp(self, duration, amplitude, f_start, f_end, t_offset=0):
        if isinstance(duration, (Register, Expression)):
            raise Q1TypeError('Chirp duration cannot be a variable or expression; '
                              'Unroll loop using Python for-loop.')
        if (isinstance(f_start, (Register, Expression))
            or isinstance(f_end, (Register, Expression))):
            raise Q1TypeError('Chirp frequency cannot be a variable or expression; '
                              'Unroll loop using Python for-loop.')

        chirp_loop_time = 100
        f_start = round(f_start)
        f_end = round(f_end)
        if f_start * f_end <= 0:
            raise Q1ValueError('Chirping through f=0.0 Hz is currently not possible')

        self.add_comment(f'chirp({duration}, {amplitude}, {f_start/1e6:7.3f}, {f_end/1e6:7.3f} MHz)')
        with self._local_timeline(t_offset=t_offset, duration=duration):
            f_step = (f_end - f_start) * chirp_loop_time / duration
            # Note: For the loop we need an integer frequency step. This could add a small
            # error of 0.5 Hz per iteration of the loop.
            # The maximum errror for a 10 ms chirp (100_000 iterations) is 0.05 MHz.
            # TODO: for more precision: add f_step fraction (use 64 bit numbers?)
            f_step = round(f_step)
            w_chirpI, w_chirpQ, delta_phase = self._waves.get_chirp(chirp_loop_time, f_step)
            # divmod, but with rem [1,100], and n > 1
            n, rem = divmod(duration-1, chirp_loop_time)
            rem += 1

            # TODO: remove workaround when fixed in firmware
            # Temporarily set NCO frequency for phase shift workaround.
            nco_freq = self.nco_frequency
            self.nco_frequency = f_start

            self.Rs._freq = int(f_start)
            self.set_gain(amplitude, amplitude)
            if n > 0:
                with self._seq_repeat(n):
                    self.shift_phase(delta_phase)
                    self.set_frequency(self.Rs._freq)
                    self.play(w_chirpI, w_chirpQ)
                    self.Rs._freq += f_step
                    self.wait(chirp_loop_time)
            self.shift_phase(delta_phase)
            self.set_frequency(self.Rs._freq)
            self.play(w_chirpI, w_chirpQ)
            self.wait(rem)
            self.set_gain(0.0)
            self.nco_frequency = nco_freq

    def _apply_paths(self, arg0, arg1):
        paths = self._enabled_paths
        if len(paths) == 0:
            raise Q1ValueError('No output paths enabled')

        if len(paths) == 1:
            path = paths[0]
            if arg1 is None:
                return (arg0, None) if path == 0 else (None, arg0)
            else:
                if path == 1:
                    # NOTE:
                    # if 1 path is enabled and 2 arguments are passed, then
                    # IQ should be rotated over pi/2, i.e. multiplied with [[0, 1], [-1, 0]]
                    # This rotation is applied to gain and offset.
                    # The amplitude of the waveform is multiplied by the gain.
                    # The waveforms only have to be swapped.
                    if isinstance(arg1, Wave):
                        return (arg1, arg0)
                    else:
                        return (-arg1, arg0)
                return (arg0, arg1)

        if paths[0] > paths[1]:
            if arg1 is None or isinstance(arg1, Wave):
                return (arg1, arg0)
            else:
                return (-arg1, arg0)

        return (arg0, arg1)

    def _translate_wave(self, wave):
        if wave is None:
            return None
        if isinstance(wave, str):
            return self._waves[wave]
        if isinstance(wave, Wave):
            return wave
        raise Q1TypeError(f'Illegal type {wave}')





from q1pulse.instrument import Q1Instrument

from init_pulsars import qcm0
from plot_util import plot_output

instrument = Q1Instrument('q1')
instrument.add_qcm(qcm0)
instrument.add_control('P1', qcm0.name, [2])
instrument.add_control('P2', qcm0.name, [3])
instrument.add_control('q1', qcm0.name, [0,1], nco_frequency=25e6)

p = instrument.new_program('loops')
p.repetitions = 3

P1 = p.P1
P2 = p.P2
q1 = p.q1

with p.loop_linspace(0.2, 1.0, 5) as amplitude:
    P1.block_pulse(20, amplitude)
    P2.block_pulse(100, amplitude - 0.1)
    q1.block_pulse(40, amplitude - 0.1)
    q1.shift_phase(0.5)

#p.describe()
#print()

p.compile(listing=True, annotate=True)

instrument.run_program(p)

plot_output([qcm0])
